@@ -1,9 +1,19 @@
+"""
+A complete bitboard representation of a chessboard, with all the methods needed
+to play a game of chess.
+"""
+
+
+try:
+    import importlib.resources as pkg_resources
+except ImportError:
+    # Python < 3.7
+    pkg_resources = None
+import random
 from copy import copy
 from math import log2
-from pathlib import Path
-import random
 
-from .moves import (
+from chessengine.moves import (
     get_white_pawn_moves,
     get_white_rook_moves,
     get_white_bishop_moves,
@@ -17,27 +27,27 @@ from .moves import (
     get_black_king_moves,
     get_black_queen_moves,
 )
-from .lookup_tables import (
+from chessengine.lookup_tables import (
     mask_position,
     clear_position,
     coords_to_pos,
     pos_to_coords,
     san_piece_map,
 )
-from .utils import get_bit_positions, get_file
-from .pgn.parser import PGNParser, SAN_MOVE_REGEX
-
-
-import logging
-
-logging.basicConfig(
-    filemode="w", filename="./log/debug_opening_book.log", level=logging.DEBUG
-)
+from chessengine.utils import get_bit_positions, get_file, clear_lines
+from chessengine.pgn.parser import PGNParser, SAN_MOVE_REGEX
 
 
 class Board:
     """
-    A class implementing a bitboard representation of a chess board
+    A class implementing a bitboard representation of a chess board.
+    A particular bitboard can be accessed via the get_bitboard method or
+    as an attribute with the name <side>_<piece>s. For example -
+
+        - white_pawns
+        - white_queens
+        - black_kings
+        - black_rooks
     """
 
     def __init__(self, side: str):
@@ -131,6 +141,8 @@ class Board:
             ("black", "knights"): "\u265E",
             ("black", "pawns"): "\u265F",
         }
+        ranks = ["8", "7", "6", "5", "4", "3", "2", "1"]
+        files = ["\u2001", "a", "b", "c", "d", "e", "f", "g", "h", "\u2001"]
 
         def add_bitboard_to_repr(board, s, p):
             board_string = bin(board)[2:]
@@ -143,9 +155,17 @@ class Board:
             add_bitboard_to_repr(self.board[(side, piece)], side, piece)
 
         board_repr = ""
+        board_repr += "\u2001" + "\u2001".join(files) + "\n"
         for i in range(8):
-            board_repr += "\u2001".join(piece_list[8 * i : 8 * i + 8][::-1])
+            board_repr += (
+                ranks[i]
+                + "\u2001\u2001"
+                + "\u2001".join(piece_list[8 * i : 8 * i + 8][::-1])
+                + "\u2001\u2001"
+                + ranks[i]
+            )
             board_repr += "\n"
+        board_repr += "\u2001" + "\u2001".join(files) + "\n"
         return board_repr
 
     def __str__(self):
@@ -157,7 +177,7 @@ class Board:
         return str(self) == str(other)
 
     def __hash__(self):
-        """I chose to just hash the bitboards instead of the FEN
+        """Just hash the bitboards instead of the FEN
         because it is faster and easier to maintain than the FEN state"""
         hash_str = f"{self.side[0]} "
         for side, piece in [
@@ -237,6 +257,9 @@ class Board:
         ]
 
     def copy(self):
+        """
+        Create and return a copy of self.
+        """
         return copy(self)
 
     def get_side_bitboard(self, side: str) -> int:
@@ -250,15 +273,18 @@ class Board:
     def get_bitboard(self, side: str, piece: str) -> int:
         """
         Returns the bitboard of the passed side for the passed pieces.
-        Calling with side="black" and piece="king" will return the black_kings bitboard, and so on.
+        For example, calling with side="black" and piece="king" will return the black_kings bitboard, and so on.
+
+        Raises AttributeError if a bitboard with an invalid name is requested. See above for the bitboard naming
+        convention.
         """
         attrname = side + "_" + piece
         return getattr(self, attrname)
 
     def get_self_piece_bitboard(self, piece: str) -> int:
         """
-        Returns the attribute corresponding to the passed piece, considering the board's
-        own side. i.e. - If the board is white, calling with piece='king' will return
+        Returns the bitboard corresponding to the passed piece, considering the board's
+        own side. i.e. - If the board is white, calling with piece="king" will return
         white king, etc.
         piece can be one of - "kings", "queens", "bishops", "knights", "rooks", "pawns"
         """
@@ -266,8 +292,9 @@ class Board:
 
     def update_board_state(self) -> None:
         """
-        Updates self.all_white, self.all_black, self.all_pieces, and self.board
-        every time a bitboard is updated
+        Updates all Board attributes when a move is made or the state of the Board
+        changes in any way. You should call this if you manually make any changes to
+        Board attributes, otherwise it is called automatically.
         """
         self.all_white = (
             self.white_pawns
@@ -314,9 +341,13 @@ class Board:
 
     def identify_piece_at(self, position: int) -> tuple:
         """
-        Identifies if there is any piece on the position passed. Returns
-        the identified piece, its side, and its board if a piece is found
-        at that position, None otherwise. Position is a power of 2
+        Identifies if there is any piece on the position passed. Position is a power of 2
+
+        Returns a 3-tuple of the format (side, piece, bitboard) where side is the side of
+        the piece identified at position (e.g, "black"), piece is the type of piece identified
+        at position (e.g, "bishops"), and bitboard is the bitboard of the piece (e.g, Board.black_bishops).
+
+        If no piece is present at position, returns (None, None, None).
         """
         for side, piece in self.board:
             board = self.board[(side, piece)]
@@ -328,6 +359,9 @@ class Board:
         """
         Moves the piece at start to end. Doesn't check anything, just makes
         the move (unless the start or end positions are invalid).
+
+        start and end are both powers of 2.
+        If track is True, stores the move in Board.moves so you can undo it later.
         """
         if not 1 <= start <= 2**63:
             raise ValueError(
@@ -370,7 +404,8 @@ class Board:
 
     def move_san(self, move: str, side: str) -> None:
         """
-        Make a move given in standard algebraic notation
+        Make a move given in standard algebraic notation.
+        move is the SAN move, side is the side to move.
         """
         # TODO - Add support for undoing castling
         if "0-0-0" in move:
@@ -440,6 +475,9 @@ class Board:
             self.move(start, end)
 
     def undo_move(self):
+        """
+        Undo the last move tracked move.
+        """
         if not self.moves:
             raise RuntimeError("No moves have been made yet to undo.")
         end, start, side, piece, board = self.moves.pop()
@@ -452,7 +490,13 @@ class Board:
         self, side: str, piece: str = None, position: int = None
     ) -> list[tuple[int, int]]:
         """
-        Gets all end positions a piece of side can reach starting from position
+        Get all end positions a piece of side can reach starting from position.
+        side is always required, piece and position are optional.
+        If piece is not specified, get all moves for all pieces of the passed side, i.e. get all valid moves for white or black.
+        If side and piece are specified and position is not, get all valid moves for the specified side and piece on
+        the board, i.e. if side is "white" and piece is "rooks", get all valid moves for all white rooks on the board.
+
+        Returns a list of moves as a list of tuples (start, end), where start and end are positions on the board.
         """
         if piece is not None:
             if position is None:
@@ -480,13 +524,24 @@ class Board:
                 return move_gens[(side, piece)](self, position)
         else:
             moves = []
-            for side, piece in self.board_pieces:
+            if side == self.side:
+                pieces = self.board_pieces
+            else:
+                pieces = self.opponent_pieces
+            for side, piece in pieces:
                 positions = get_bit_positions(self.get_bitboard(side, piece))
                 for position in positions:
                     moves.extend(self.get_moves(side, piece, position))
             return moves
 
     def search_forward(self, depth: int = 4) -> tuple[int, tuple]:
+        """
+        Execute an alpha-beta pruned depth-first search to find the optimal move from
+        the current board state.
+
+        Arguments -
+        depth: int - The number of plies to search (1 move is 2 plies). Default = 4 plies.
+        """
         maximize = self.side == "white"
         best_score = -1000 if maximize else 1000
 
@@ -515,6 +570,17 @@ class Board:
         beta: int = 1000,
         maximizing_player: bool = True,
     ) -> int:
+        """
+        Execute an alpha-beta pruned search. You probably won't need to
+        call this function yourself, use Board.search_forward instead.
+
+        :param depth: The number of plies to search forward (default=4)
+        :param alpha: The minimum score that the maximizing player is guaranteed (default=-1000). You probably won't need to specify this argument.
+        :param beta: The maximum score that the minimizing player is guaranteed (default=1000). You probably won't need to specify this argument.
+        :param maximizing_player: True if white is searching for a move, False if black is searching for a move.
+
+        Returns the value of the best board position found.
+        """
         if depth == 0:
             return self.score
 
@@ -547,84 +613,135 @@ class Board:
         """
         The game loop.
         """
+        parser = None
+        if pkg_resources is not None:
+            loading_messages = [
+                "Searching for opening moves.",
+                "Reading an opening book.",
+                "Waking up.",
+                "Reading thousands of past games.",
+            ]
 
-        def clear_lines(n):
-            """
-            Clears the last n lines printed so we can print there again
-            """
-            LINE_UP = "\033[1A"
-            LINE_CLEAR = "\x1b[2K"
-            for i in range(n):
-                print(LINE_UP, end=LINE_CLEAR)
+            parser = PGNParser()
+            print(random.choice(loading_messages))
+            opening_files = pkg_resources.contents("chessengine.openings")
+            for child in opening_files:
+                file_path = pkg_resources.path("chessengine.openings", child)
+                with file_path as f:
+                    if f.suffix == ".pgn":
+                        print(".")
+                        parser.parse(f)
+            print(f"\nRead through {len(parser.games)} games.")
+        print(f"Set search depth to {search_depth}")
 
-        loading_messages = [
-            "Searching for opening moves.",
-            "Reading an opening book.",
-            "Waking up.",
-            "Reading thousands of past games."
-        ]
-
-        parser = PGNParser()
-        print(random.choice(loading_messages))
-        opening_files = Path("./chessengine/openings")
-        for child in opening_files.rglob("*.pgn"):
-            print('.')
-            parser.parse(child)
-        print(f"\nRead through {len(parser.games)} games.")
-        print(f'Set search depth to {search_depth}')
-
-        print("\n" * 11)
+        print(self)
         side_to_move = "white"
-        current_node = parser.root_node
-        in_game_tree = True
+        in_game_tree = parser is not None
+        current_node = parser.root_node if in_game_tree else None
+        lines_printed = 11
+        last_move = ""
         while True:
+            clear_lines(lines_printed)
+            print(self)
+            lines_printed = 11
             if side_to_move == self.side:
-                clear_lines(11)
                 if in_game_tree:
                     move, node = random.choice(list(current_node.children.items()))
-                    logging.debug(
-                        f"Chose to make move {move} out of {current_node.children.keys()}"
-                    )
                     self.move_san(move=move, side=side_to_move)
                     current_node = node
-                    print(f"Board moves {move}")
+                    last_move = f"Board moves {move}"
                 else:
                     best_score, best_move = self.search_forward(search_depth)
-                    logging.debug(
-                        f"Chose to make move {pos_to_coords[log2(best_move[0])]} to {pos_to_coords[log2(best_move[1])]}"
-                    )
                     self.move(best_move[0], best_move[1])
-                    print(
-                        f"Board moves from {pos_to_coords[log2(best_move[0])]} to {pos_to_coords[log2(best_move[1])]}"
-                    )
-                print(self)
+                    last_move = f"Board moves from {pos_to_coords[log2(best_move[0])]} to {pos_to_coords[log2(best_move[1])]}"
             else:
-                move = input(
-                    "Enter the move you want to make in standard algebraic notation - "
-                ).strip()
-                if move.lower() == "q":
-                    print(f"Thanks for playing!")
-                    return
-                if move.lower() == "u":
+                # ask for user input until accepted:
+                input_accepted = False
+                move_undone = False
+                move = ""
+                while not input_accepted:
+                    print(last_move)
+                    move = input(
+                        "Enter the move you want to make in standard algebraic notation - "
+                    ).strip()
+                    lines_printed += 2
+                    if move.lower() == "q":
+                        print("Thanks for playing!")
+                        return
+                    if move.lower() == "u":
+                        move_undone = True  # this helps us break out of the input loop
+                        try:
+                            self.undo_move()
+                            self.undo_move()
+                        except RuntimeError:
+                            print("No moves have been made yet to undo!")
+                            lines_printed += 1
+                        break
+
+                    # input was normal move
                     try:
-                        self.undo_move()
-                        self.undo_move()
-                    except RuntimeError:
-                        print("No moves have been made yet to undo!\n")
+                        self.move_san(move=move, side=side_to_move)
+                        input_accepted = True
+                    except ValueError as e:
+                        print(e)
+                        lines_printed += 1
+                if move_undone:
+                    # return to outer loop, so both sides need to make a new move
                     continue
 
-                logging.debug(f"Player made move {move}")
-                self.move_san(move=move, side=side_to_move)
                 if in_game_tree:
                     try:
                         current_node = current_node.get_child(move)
-                        logging.debug(f"Still in game tree")
                     except ValueError:
-                        logging.debug("Moved out of game tree")
                         in_game_tree = False
-                clear_lines(11)
-                print(f"You moved from {move[0]} to {move[1]}")
-                print(self)
+                print(f"You moved {move}")
+                lines_printed += 1
+
+            if side_to_move == "white":
+                side_to_move = "black"
+            else:
+                side_to_move = "white"
+
+    def play_pvp(self) -> None:
+        """
+        Play a game of chess on the terminal with another player
+        """
+        print(self)
+        lines_printed = 11
+
+        side_to_move = "white"
+        last_move = ""
+        while True:
+            clear_lines(lines_printed)
+            print(self)
+            lines_printed = 11
+
+            input_accepted = False
+            while not input_accepted:
+                print(last_move)
+                move = input(
+                    f"Enter the move you want to make ({side_to_move.capitalize()}'s turn) - "
+                ).strip()
+                lines_printed += 2
+
+                if move.lower() == "q":
+                    print("Thanks for playing!")
+                    return
+                if move.lower() == "u":
+                    try:
+                        self.undo_move()
+                    except RuntimeError:
+                        print("No moves have been made yet to undo!")
+                        lines_printed += 1
+                    break
+
+                try:
+                    self.move_san(move=move, side=side_to_move)
+                    input_accepted = True
+                    last_move = f"{side_to_move.capitalize()} moved {move}"
+                except ValueError as e:
+                    print(e)
+                    lines_printed += 1
 
             if side_to_move == "white":
                 side_to_move = "black"
